/*
 * Copyright (C) 2018 Richard Hughes <richard@hughsie.com>
 *
 * SPDX-License-Identifier: LGPL-2.1+
 */

#pragma once

#define __LIBJCAT_H_INSIDE__

#include <libjcat/jcat-blob.h>
#include <libjcat/jcat-common.h>
#include <libjcat/jcat-context.h>
#include <libjcat/jcat-engine.h>
#include <libjcat/jcat-file.h>
#include <libjcat/jcat-item.h>
#include <libjcat/jcat-result.h>
#include <libjcat/jcat-version.h>

#undef __LIBJCAT_H_INSIDE__
